	processor 6502
	include vcs.h
	include macro.h

	;constants
PAL		equ	1
DEBUG_SCORE	equ	0
SHOW_PADDLE	equ	0	;if true, show paddle value in score display

#if PAL

;PAL
VBLNK	equ	48
LINES	equ	228
OVERSCN	equ	36
FPS		equ	50

#else

;NTSC
VBLNK	equ	40
LINES	equ	192
OVERSCN	equ	30
FPS		equ	60

#endif

VBLNK64	equ	VBLNK*19/16		;value to set TIM64 to for VBLNK
OVERS64 equ OVERSCN*19/16

SLINES	equ	18				;number of lines used by score
SCORESZ	equ	7				;number of bytes used for score. 6 B mantissa + 1 B exponent (exponent is MSB and thus at +4)
SCORED	equ	12				;number of digits in score
BLINES	equ	4				;bottom lines

KERNEL	equ	2				;size of kernel
KLINES	equ	(LINES-SLINES-BLINES)/KERNEL	;number of kernel lines
BELLH	equ	8				;height of bell in kernel lines
BELLW	equ	16				;width of bell in pixels
RABBITH	equ	8				;height of rabbit in kernel lines
RABBITW	equ	8				;width of rabbit in pixels
GAP		equ	12				;size of gap in kernel lines. must be >= 2
BACKGH	equ	24				;height of background in kernel lines
NBELLS	equ	15
BRDFREQ	equ	10				;how often birds appear
BRDSPED	equ	1				;bird speed

XMIN	equ 10				;constraint player within these X values
XMAX	equ	150
BXMIN	equ 20				;constraint bells within these X values
BXMAX	equ	132
YMIN	equ	10				;constraint rabbit within these Y values
YMAX	equ	YMIN+KLINES/2
XDESTV	equ	4				;the speed at which xdest moves
#if PAL
JUMPVEL	equ	525				;in 8.8 fixed point units
GJMPVEL	equ	650				;like JUMPVEL, but used when on the ground
GRAVITY	equ -17				;ditto
#else
JUMPVEL	equ	435				;PAL JUMPVEL * 5/6, slightly adjusted
GJMPVEL	equ	500				;adjusted so the camera doesn't scroll up when reaching the peak of the jump
GRAVITY	equ -12				;PAL GRAVITY * (5/6)^2
#endif
FALLMIN	equ	-20				;we can't fall faster than this. should be >= -BELLH-GAP
HOPVEL	equ	175				;speed when hopping on ground
HOPFRMS	equ	6				;number of hop frames
SITFRMS	equ	4				;number of sit frames
FBELL	equ	1				;first touchable bell
VBELLS	equ	(KLINES+GAP+BELLH-1)/(GAP+BELLH)	;max visible bells
TREEH	equ	12				;height of the trees
HOLD	equ	100				;how many frame to hold each line for

;colors
BACKCLR	equ $00
PLAYCLR	equ	$0E			;player
BELLCLR	equ	$0E			;bell
SCORCLR	equ	$0C			;score
HISCCLR	equ	$2E			;hiscore
TEXTCLR	equ	$0C			;normal text

	;RAM
	SEG.U VARS
	org $80
xpos	ds	2
xdest	ds	1
paddle	ds	1
ypos	ds	2
yvel	ds	2
level	ds	2			;signed height above ground. kept within [0,$7FFFF]
						;ground is only visible if level+1 == 0
score	ds	SCORESZ
hiscore	ds	SCORESZ
bellscr	ds	2			;current score per bell, in BCD
scorept	ds	2*SCORED	;pointers to score sprites
						;TODO: we can halve this since we're only drawing
						;      half the sprites each frame
cenam1	ds	1			;current ENAM1, derived from exponent (decimal point)
scorey	ds	1			;Y position of score, [-7,7]
						;<0: crop top, stuff lines after
						; 0: normal
						;>0: crop bottom, stuff lines before
scoreyb	ds	1			;post score kernel burn, scoreyb = max(0,-scorey)
curline	ds	1			;current line shown in display
						;0-(TEXTLINES-1) = show that line
						;NOHISCRLINE = NOHISCRLINE = show text declaring the player a loser
						;HISCORELINE = NOHISCRLINE+1 = show text declaring new hiscore
						;HISCORELINE+1 = show hiscore
						;HISCORELINE+2 = show score
						;alternate between the last two
holdcnt	ds	1			;how many frames the current line has been held
textclr	ds	1			;color of the text
scorcnt	ds	1			;how many times the score has been shown
						;once >= SCORCNTMAX, set curline = 0
SCORCNTMAX	equ	10

;pointers
p0ptr	ds	2			;pointer to P0 sprite data

;counters
rabcntr	ds	1			;rabbit counter
belcntr	ds	1			;bell/bird counter
bakcntr	ds	1			;background counter
brdcntr	ds	1			;bird counter
						;< 0: spawn after -brdcntr/2 bells
						;>= 0: pointer into bells array
						;never >= 2*NBELLS

;each bell has the following as its state:
;byte	xpos			;horizontal position
;byte	frame_state		;current animation frame and visibility state
; b6-7	sprite			;0 = invisible (black), 1 = explosion, 2 = bell, 3 = bird
; b3-5	frame			;current animation frame
; b0-2	subframe		;low order bits of current frame (divide by eight)
bells	ds	2*NBELLS
belstrt	ds	1			;starting line of current bell - for smooth scrolling
						;if > BELLH, skip belstrt-BELLH kernel lines before drawing the first bell
belshft	ds	1			;number of lines to shift bells down

linleft	ds	1			;numer of kernel lines left in frame
frame	ds	1			;used for animating the bells
rabbref	ds	1			;REFP0

insptr0	ds	2			;pointer to instrument 0
insptr1	ds	2			;pointer to instrument 1
inscnt0 ds	1			;offset into instptr0
inscnt1 ds	1			;offset into instptr1

;General game/player state
; b7	player is off the ground and has touched a bell (aka "the game is on")
; b6	player is off the ground
; b3	bird direction. also REFP1
state	ds	1

rand	ds	2			;prng state
temp	ds	2

#if !PAL
mod6sum ds  1           ;for computing frame mod 6
#endif

	echo "RAM:", ($100 - *), "bytes left"

	;ROM
	SEG CODE

	org $F000
	.byte 0
	align 64
	include graphics.h
	echo "Page $F0:", ($F140 - *), " bytes left"
	align 1

	org $F100
	include rabbitgfx.h
	include scorekernel.h
	echo "Page $F1:", ($F200 - *), " bytes left"

	org $F200
NUMBER_PAGE	equ	>NumberZero
	include scoregfx.h
	include background.h
	echo "Page $F2:", ($F300 - *), " bytes left"

	org $F300
	include mainkernel.h
	include font.h
 
	;Font should stay within one page
	echo "Font: [", Font, ",", *, ") -> page difference:", (>Font - >*)

	include text.asm
	include soundfx.h

#if !PAL
Mod6Tab1    ;x mod 6, 0 <= x <= 13 (5 + 4 + 4)
    .byte 0
    .byte 1
    .byte 2
    .byte 3
    .byte 4
    .byte 5
    .byte 0
    .byte 1
    .byte 2
    .byte 3
    .byte 4
    .byte 5
    .byte 0
    .byte 1
Mod6Tab2    ;2x mod 6, 0 <= x <= 7
    .byte 0
    .byte 2
    .byte 4
    .byte 0
    .byte 2
    .byte 4
    .byte 0
    .byte 2
Mod6Tab4    ;4x mod 6, 0 <= x <= 3
    .byte 0
    .byte 4
    .byte 2
    .byte 0
#endif

Start
	CLEAN_START

	;RAM and all TIA registers are zero
	;Note that we used to set the following TIA registers manually:
	; SWACNT = 0	joystick = input
	; CTRLPF = 0	repeating playfield

	;put player in center
	lda #(XMIN+XMAX)/2
	sta xpos+1
	sta xdest

	lda #YMIN
	sta ypos+1

	;set up P0 and P1 pointer high bytes
	lda #>Rabbit0
	sta p0ptr+1

	lda #<-7
	sta scorey
#if !PAL
    jmp OverscanLogic
#endif

MainLoop
	VERTICAL_SYNC
	lda #VBLNK64
	sta TIM64T

;************************
;** do main logic here **
;************************

    ;Turning this into a macro so we can paste it into the overscan part for NTSC
    MAC OVERSCAN_LOGIC
	inc frame

#if !PAL
    ;For our hackish NTSC support we simply skip certain parts of the logic every 6th frame.
    ;Hence we must compute frame mod 6 and check that it's non-zero in strategic places.
    ;Splitting frame up into two three-bit parts and one two-bit part like frame = a + 8b + 64c,
    ;we can use the fact that:
    ; frame mod 6 = (a mod 6) + (8b mod 6) + (64c mod 6) mod 6 = (a mod 6) + (2b mod 6) + (4c mod 6) mod 6
    lda frame
    and #%111
    tax
    lda Mod6Tab1,X
    sta mod6sum     ;mod6sum = a mod 6
    lda frame
    lsr
    lsr
    asr #%1110      ;carry clear
    tax
    lda Mod6Tab2,X
    adc mod6sum
    sta mod6sum     ;mod6sum = (a mod 6) + (2b mod 6)
    lda frame
    rol
    rol
    rol
    and #%11
    tax
    lda Mod6Tab4,X  ;A = 4c mod 6
    clc
    adc mod6sum
    tax             ;X = (a mod 6) + (2b mod 6) + (4c mod 6)
    lda Mod6Tab1,X  ;0 <= X <= 5 + 4 + 4 = 13
    sta mod6sum     ;mod6sum = frame mod 6
#endif

	;twiddle prng state
	lda frame
	adc rand
	adc xpos
	sta rand
	adc xpos+1
	adc rand+1
	sta rand+1
	
	;apply gravity to velocity
	clc
	lda yvel
	adc #<GRAVITY
	sta yvel
	lda yvel+1
	adc #>GRAVITY
	sta yvel+1

	;if color, then paddles
	;if B/W, then joystick
	lda SWCHB
	and #$08
	beq DoJoystick

	;paddles
#if SHOW_PADDLE
	sed
	lda #0
	sta score+1
	ldx paddle
	clc
ToBCD
	adc #1
	bcc Next
	inc score+1
Next
	dex
	bne ToBCD
	sta score
	cld
#endif

	;Adjust paddle value
	;It's something like 15-130 when we enter here.
	;We want it to cover the [XMIN,XMAX] range.
	;This means we need to double the value and subtract
	;a value to move everything "to the left" and widen the range.
	;Care must also be taken that we don't over- or underflow
	;the accumulator.
ADJUST	equ	30

	lda paddle
	bmi ClampPaddleRight	;ASL would overflow - clamp to XMAX
	asl
	cmp #XMIN+ADJUST
	bcc ClampPaddleLeft		;subtracting ADJUST would net us A < XMIN - clamp
	sec
	sbc #ADJUST
	cmp #XMAX
	bcc DoneClampingXDest	;accumulator in range
ClampPaddleRight
	lda #XMAX
	.byte $2C
ClampPaddleLeft
	lda #XMIN
DoneClampingXDest
	sta xdest
	jmp DoneController

DoJoystick
	;process joystick input
	;left?
	lda #$40
	bit SWCHA
	bne DoneLeft
	
	;move xdest left
	lda xdest
	sec
	sbc #XDESTV
	cmp #XMIN
	bcc ClampLeft
	
	sta xdest
	jmp DoneLeft
	
ClampLeft
	lda #XMIN
	sta xdest
	
DoneLeft
	;right?
	lda #$80
	bit SWCHA
	bne DoneRight
	
	;move xdest left
	lda xdest
	clc
	adc #XDESTV
	cmp #XMAX
	bcs ClampRight
	
	sta xdest
	jmp DoneRight
	
ClampRight
	lda #XMAX
	sta xdest
	
DoneRight
DoneController
	;check whether we're on the ground
	lda level
	ora level+1
	bne NotOnGround
	lda ypos+1
	cmp #YMIN
	bne NotOnGround

	lda state
	ora #$40
	sta state

	;check whether the game was already started
	;if so, end it
	bpl CheckJumpWithButton
	jsr EndGame
	jmp NotJumping

CheckJumpWithButton
	;check whether the player is pressing the button
	;if so, jump
	;check SWCHA/INPT4 depending on SWCHB
	;we used to AND them, but that made pushing the joystick right jump
	lda SWCHB
	and #$08
	beq CheckJoystickButton

	;paddle button
	lda SWCHA
	.byte $0C			;NOP Abs instead of BIT Abs since we need the flags
CheckJoystickButton
	lda INPT4
	bmi NotJumpingWithButton
	jmp DoJump

NotOnGround
	lda state
	and #<~$40
	sta state

NotJumpingWithButton
	;make sure we've cleared the trees
	lda level+1
	bne ClearedTrees
	lda level
	clc
	adc ypos+1
	bcs ClearedTrees
	cmp #TREEH+YMIN
	bcs ClearedTrees
	jmp NotJumping

ClearedTrees
	;check whether we touched a bell
	lda #$80
	bit CXPPMM

	;a beq NotJumping would be better here, but we can't reach that label with a branch
	bne P0P1Touched
	jmp NotJumping

P0P1Touched
	;figure out which bell we touched by doing a coarse check of the bounding boxes
	;the code below is a little hackish to handle some corner cases, but seems to work well
	;temp, aka y, holds the y position of the current bell
	lda #KLINES-FBELL*(GAP+BELLH)
	sec
	sbc belstrt
	clc
	adc #BELLH
	sta temp
	ldx #2*FBELL

CheckBoundingBoxes
	;xpos > x+BELLW+4?
	lda bells,X
	clc
	adc #BELLW+4
	cmp xpos+1
	bmi NotOverlapping

	;xpos+RABBITW+4 < x?
	lda xpos+1
	clc
	adc #RABBITW+4
	cmp bells,X
	bmi NotOverlapping

	;ypos-RABBITH >= y?
	lda ypos+1
	sec
	sbc #RABBITH
	cmp temp
	bpl NotOverlapping

	;y-BELLH-2 >= ypos?
	lda temp
	sec
	sbc #BELLH+2
	cmp ypos+1
	bpl NotOverlapping

	;bounding boxes overlap
	;make sure this isn't an explosion or invisible
	lda bells+1,X
	bpl NotJumping

	;save X since ScoreBell and ScoreBird destroy it
	;temp is safe since it's used by neither function
	stx temp

	;we hit something
	;call StartGame if game hasn't started yet
	lda state
	bmi GameAlreadyStarted
	jsr StartGame

	;Note: X comes out as zero from StartGame but the first
	;      thing hit is never the bird since it takes some
	;      time to appear. Hence the code below never branches
	;      to DoScoreBird

GameAlreadyStarted
	;score. bird or bell based on bit 6
	lda bells+1,X
	and #$40
	bne DoScoreBird

	;play bell sound
	jsr ScoreBell
	ldy #0
	beq DoneScoring

DoScoreBird
	;reset brdcntr. this also stops the explosion from moving and the bells from reflecting
	lda #<(-2*BRDFREQ)
	sta brdcntr

	;play bird sound
	jsr ScoreBird
	ldy #2

DoneScoring
	jsr LoadFreeChannel

	;explode
	lda #$40
	ldx temp
	sta bells+1,X

	jmp DoJump

NotOverlapping
	lda temp
	sec
	sbc #BELLH+#GAP
	sta temp
	inx
	inx
	cpx #VBELLS*2
	bne CheckBoundingBoxes

	;if we got here then no bounding boxes overlap and the above code is a little buggy (or the TIA is)
	;avoid false positives by pretending we didn't collide
	;hopefully we'll get it next frame, or we pretend it was a near miss
	jmp NotJumping

DoJump
	;jump
	;reset frame and velocity
	lda state
	bmi JumpNormal

	;on ground
	lda #<GJMPVEL
	sta yvel
	lda #>GJMPVEL
	sta yvel+1
	jmp NotJumping

JumpNormal
	lda #<JUMPVEL
	sta yvel
	lda #>JUMPVEL
	sta yvel+1

NotJumping
    ENDM

#if PAL
    OVERSCAN_LOGIC
#endif

	sta CXCLR
	
	jsr HandleMusic

	;keep yvel+1 >= FALLMIN
	lda yvel+1
	bpl NotFallingTooFast	;yvel+1 >= 0
	cmp #<FALLMIN
	bpl NotFallingTooFast	;yvel+1 >= FALLMIN
	lda #<FALLMIN
	sta yvel+1

NotFallingTooFast
	;apply velocity
	clc
	lda ypos
	adc yvel
	sta ypos
	lda ypos+1
	adc yvel+1
	sta ypos+1

	;calculate belshft based on rabbit's Y position and the fall speed
	;use temp to save camera movement, which is sign extended and added to level futher down
	lda #0
	sta temp
	lda frame
	and #3
	bne NoBelShft	;down once every four framess
#if !PAL
    lda mod6sum
    beq NoBelShft   ;down somewhat slower in NTSC (just under ever fifth frame)
                    ;actually, slightly faster since there are times when 0 == frame mod 4 == frame mod 6
#endif
	lda #1
	.byte $2C
NoBelShft
	lda #0
	sta belshft
	
	;keep ypos+1 within [YMIN, YMAX)
	lda ypos+1
	sec
	sbc #YMIN
	bpl NotTooFarDown			;ypos+1 >= YMIN
	sta temp

	lda #YMIN
	sta ypos+1

	;only shift bells if level > 0, meaning if we're off the ground
	lda level
	bne DoPerformYMIN
	lda level+1
	beq DoneConstraintingYpos

DoPerformYMIN
	lda temp
	clc
	adc belshft
	sta belshft
	
	jmp DoneConstraintingYpos
NotTooFarDown
	lda ypos+1
	sec
	sbc #YMAX
	bmi DoneConstraintingYpos	;ypos+1 < YMAX
	sta temp
	clc							;guaranteed?
	adc belshft
	sta belshft
	lda #YMAX
	sta ypos+1

DoneConstraintingYpos
	;modify level according to temp
	;keep it within [0,$7FFF] though
	lda temp
	bmi TempIsMinus

	;temp is positive
	;don't add to level if it's already >= $7F00
	ldx level+1
	cpx #$7F
	bcs LevelModificationDone

	adc level
	sta level
	bcc LevelModificationDone
	inc level+1
	bne LevelModificationDone

TempIsMinus
	clc
	adc level
	sta level
	bcs LevelModificationDone
	dec level+1

	;if(level < 0) level = 0
	bpl LevelModificationDone
	lda #0
	sta level
	sta level+1

LevelModificationDone
	jsr ShiftBells

	;move bird if visible
	ldx brdcntr
	bmi NotMovingBird
#if !PAL
    lda mod6sum
    beq NotMovingBird   ;slow bird down slightly in NTSC
#endif
	lda state
	and #$08
	beq MoveBirdRight

	;moving bird left
	lda bells,X
	sec
	sbc #BRDSPED
	cmp #BXMIN
	bcs MoveBirdDone

	;clamp x position and reverse direction
	lda #BXMIN
	sta bells,X
	lda state
	and #<~$08
	sta state
	jmp NotMovingBird

MoveBirdRight
	lda bells,X
	clc
	adc #BRDSPED
	cmp #BXMAX
	bcc MoveBirdDone
	
	;clamp x position and reverse direction
	lda #BXMAX
	sta bells,X
	lda state
	ora #$08
	sta state
	jmp NotMovingBird	

MoveBirdDone
	sta bells,X

NotMovingBird
	;move horizontaly
	lda state
	and #$40
	beq MoveExponentially

	ldx xpos
	ldy xpos+1
	cpy xdest
	beq DoPickSitFrame
	bcc MoveRight
	bcs MoveLeft

DoPickSitFrame
	jsr PickSitFrame
	jmp DoneReflecting

MoveLeft
	jsr PickHopFrame

	;move left (carry is set)
	txa					;xpos
	sbc #<HOPVEL
	sta xpos
	tya					;xpos+1
	sbc #>HOPVEL
	sta xpos+1

	;make sure we didn't overshoot xdest
	cmp xdest
	bcs NotReflecting
	lda xdest
	sta xpos+1
	bne NotReflecting	;xdest is never zero

MoveRight
	jsr PickHopFrame

	;move right (carry is clear)
	txa					;xpos
	adc #<HOPVEL
	sta xpos
	tya					;xpos+1
	adc #>HOPVEL
	sta xpos+1

	;make sure we didn't overshoot xdest
	cmp xdest
	bcc Reflecting
	lda xdest
	sta xpos+1
	bne Reflecting		;xdest is never zero

MoveExponentially
	jsr PickJumpFrame

#if !PAL
    ;HACKHACK: This is really ugly, but adapting the exponential movement
    ;          below with another factor is simply not reasonable.
    lda mod6sum
    beq SkipHorzontalMovement
#endif

	;exponentially approach xdest
	;xpos = xpos - (xpos >> 3) + (xdest >> 3)
	;FIXME: this looks awfully long for such a simple operation
	lda xpos
	lsr
	lsr
	lsr
	sta temp
	lda xpos+1
	asl
	asl
	asl
	ora temp
	sta temp
	lda xpos+1
	lsr
	lsr
	lsr
	sta temp+1
	sec
	lda xpos
	sbc temp
	sta xpos
	lda xpos+1
	sbc temp+1
	sta xpos+1
	lda xdest
	asl
	asl
	asl
	sta temp
	lda xdest
	lsr
	lsr
	lsr
	sta temp+1
	clc
	lda xpos
	adc temp
	sta xpos
	lda xpos+1
	adc temp+1
	sta xpos+1

	;don't touch rabbref if xpos+1 == xdest
	cmp xdest
	beq DoneReflecting

	;check if we should reflect the rabbit
	lda xpos+1
	cmp xdest
	bcs NotReflecting
Reflecting
	lda #0
	.byte $2C
NotReflecting
	lda #$08
	sta rabbref
DoneReflecting
SkipHorzontalMovement
	ldx #NBELLS*2
StepBellFrames
	;first check if this is an explosion
	lda bells-1,X
	and #$C0
	beq StepBellFrames_Next			;skip blacks
	cmp #$40
	beq StepBellFrames_Explosion

	lda bells-1,X
	and #$3F
	clc
	adc #1
	and #$3F
	sta temp
	lda bells-1,X
	and #$C0
	ora temp
	sta bells-1,X
	jmp StepBellFrames_Next

StepBellFrames_Explosion
	lda bells-1,X
	and #$3F
	clc
	adc #4		;animation speed
	cmp #64
	bpl StepBellFrames_Explosion_GoBlack
	ora #$40
	.byte $2C
StepBellFrames_Explosion_GoBlack
	;end of explosion - go black
	lda #0
	sta bells-1,X

StepBellFrames_Next
	dex
	dex
	bne StepBellFrames

	;if the game has started, then we should only draw the score
	;no scrolling or anything
	lda state
	bpl Text_DoScroll

	;FIXME: Since we enter ShowScore with scorey being either 0 or 7
	;we need to add scorey to all pointers.
	;There might be a way to not have to mess with the pointers at all,
	;but that would require modifying the score kernel a bit.
	;Considering how short on time we are in that kernel this
	;is probably not possible though.

	lda #0
	sta scorey
	jmp ShowScore

Text_DoScroll
	;handle text
	ldx scorey
	beq Text_Hold		;if scorey == 0, then hold for a while

	;scroll up every fourth frame
	lda frame
	and #3
	beq Text_ScrollUp

	;nothing to do this frame
	bne Text_JMP_Text_End

Text_Hold
	;hold the current line of text for some time
	ldx holdcnt
	inx
	cpx #HOLD
	bcs Text_StopHold
	stx holdcnt
	bne Text_JMP_Text_End

Text_StopHold
	;fun's over - start scrolling up
	dec scorey

Text_JMP_Text_End
	jmp Text_End

Text_ScrollUp
	dex
	stx scorey
	cpx #<-8
	beq Text_LoadNext

	;if scorey is positive, adjust scorept:s
	cpx #0
	bmi Text_JMP_Text_End

Text_ScrollUpPointers
	ldx #2*SCORED
Text_ScrollUpPointers_Loop
	dec scorept-2,X
	dex
	dex
	bne Text_ScrollUpPointers_Loop
	beq Text_JMP_Text_End

Text_LoadNext
	lda #7
	sta scorey
	lda #0
	sta holdcnt

#if DEBUG_SCORE
	lda #$12
	sta score+5
	lda #$34
	sta score+4
	lda #$56
	sta score+3
	lda #$78
	sta score+2
	lda #$90
	sta score+1
	lda #$12
	sta score+0
	lda #$56
	sta score+6
	jmp ShowScore
#endif
#if SHOW_PADDLE
	jmp ShowScore
#endif

	lda curline
	inc curline
	cmp #HISCORELINE+1
	bcc LoadText
	cmp #HISCORELINE+2
	beq ShowScore
	cmp #HISCORELINE+1
	bcs ShowHiscore			;>= HISCORELINE+1

LoadText
	cmp #TEXTLINES-1
	beq LoadText_SkipToScore
	cmp #NOHISCRLINE
	bne LoadText_Normal

LoadText_SkipToScore
	;we need to follow TEXTLINES-1 and NOHISCRLINE with HISCORELINE+1
	ldx #HISCORELINE+1
	stx curline

LoadText_Normal
	;load A:th line of text
	;Y = A*12
	asl
	asl
	sta temp
	asl
	adc temp			;carry is clear
	tay
	ldx #0

	lda #TEXTCLR
	sta textclr

LoadText_Loop
	lda Text,Y
	cmp #64
	bcs LoadText_NotSpace
	cmp #32				;' '
	beq LoadText_Space

	;digit
	sec
	sbc #48				;'0'
	asl
	asl
	asl
	adc #<NumberZero+7		;scorey is always 7
	sta scorept,X
	lda #>NumberZero
	bne LoadText_Next

LoadText_Space
	lda #<NumberBlank+7
	sta scorept,X
	lda #>NumberBlank
	bne LoadText_Next

LoadText_NotSpace
	sec
	sbc #65				;'A'
	asl
	asl
	asl
	adc #<Font+7		;scorey is always 7
	sta scorept,X
	lda #>Font

LoadText_Next
	sta scorept+1,X
	iny
	inx
	inx
	cpx #SCORED*2
	bne LoadText_Loop
	jmp Text_End

ShowHiscore
	;curline might be HISCORELINE+3
	lda #HISCORELINE+2
	sta curline

	lda #HISCCLR
	sta textclr

	;transform hiscore to scorept
	lda #<(hiscore-1)
	sta temp
	lda #>(hiscore-1)
	sta temp+1
	jmp TransformScore

ShowScore
	lda #SCORCLR
	sta textclr

	inc scorcnt
	lda scorcnt
	cmp #SCORCNTMAX
	bcc ShowScore_Normal

	;go back to showing line zero after this
	lda #0
	sta curline
	sta scorcnt

ShowScore_Normal
	lda #<(score-1)
	sta temp
	lda #>(score-1)
	sta temp+1

TransformScore
	;set up score pointer high bytes
	;set low bytes as well - cheaper on ROM and they get set later anyway
	lda #NUMBER_PAGE
	ldx #2*SCORED
NumberPtrLoop
	sta scorept-1,X
	dex
	bne NumberPtrLoop

	ldy #SCORESZ-1
	ldx #0

	;make leading zeroes blanks
	clc
TransformScore_ZeroLoop
	lda (temp),Y
	bne TransformScore_NonZero
	lda #<NumberBlank
	adc scorey
	sta scorept,X
	sta scorept+2,X
	inx
	inx
	inx
	inx
	dey
	bne TransformScore_ZeroLoop

	;that was the last digit, meaning the score was zero
	;put a single zero in the rightmost place and be done
	lda #<NumberZero
	adc scorey
	sta scorept+2*SCORED-2
	jmp ExpIsZero

TransformScore_NonZero
	cmp #$10
	bcc TransformScore_LeftDigit

TransformScore_Loop
	lda (temp),Y
	asr #$F0
	adc #<NumberZero
	.byte $2C
TransformScore_LeftDigit
	;only the left digit in this pair is blank
	lda #<NumberBlank
	adc scorey
	sta scorept,X
	inx
	inx
	lda (temp),Y
	and #$0F
	asl
	asl
	asl
	adc #<NumberZero
	adc scorey
	sta scorept,X
	inx
	inx
	dey
	bne TransformScore_Loop

	;check whether exp > 0
	ldy #SCORESZ
	lda (temp),Y
	beq ExpIsZero

	lda #<NumberE
	adc scorey
	sta scorept+18
	lda (temp),Y
	asr #$F0
	adc #<NumberZero
	adc scorey
	sta scorept+20
	lda (temp),Y
	and #$0F
	asl
	asl
	asl
	adc #<NumberZero
	adc scorey
	sta scorept+22
	lda #2
	.byte $2C
ExpIsZero
	lda #0
	sta cenam1

Text_End
	lda #0		;yes, we really need to do this. the above LDA above is part of a BIT trick
	sta GRP0
	sta GRP1
	sta PF0
	sta PF1
	sta PF2

	;set up counters
	sec
	lda #KLINES+RABBITH
	sbc ypos+1
	sta rabcntr
	lda belstrt
	sta belcntr

	lda level+1
	bne DontShowGround

	lda #KLINES-2
	clc
	adc level
	bcs DontShowGround
	bmi DontShowGround
	sec
	sbc #BACKGH
	.byte $2C
DontShowGround
	lda #$7F
	sta bakcntr

	lda #BACKCLR
	sta COLUBK

	;position P0 and P1 for score display
SCOREXPOS	equ	50
	lda frame
	and #1
	beq PositionP0Even
	lda #7
PositionP0Even
	clc
	adc #SCOREXPOS
	ldx #0
	jsr PositionObject
	lda frame
	and #1
	beq PositionP1Even
	lda #7
PositionP1Even
	clc
	adc #SCOREXPOS+16
	ldx #1
	jsr PositionObject

	;position M0 (decimal point)
	lda #SCOREXPOS+7
	ldx #3
	jsr PositionObject

	lda xdest
	clc
	adc #RABBITW/2
	ldx #2
	jsr PositionObject

	lda #6
	;three copies, medium
	sta NUSIZ0
	sta NUSIZ1

	;no reflection
	lda #0
	sta REFP0
	sta REFP1
	sta scoreyb

	lda frame
	bmi StarYellow

	;darker yellow
	lda #$26
	.byte $2C
StarYellow
	lda #$2E
	sta COLUPF

	lda textclr
	sta COLUP0
	sta COLUP1

WaitForVblankEnd
	lda INTIM
	bne WaitForVblankEnd
	lda #0
	sta VBLANK

	;WSYNC to avoid post-VBLANK jitter
	sta WSYNC
	sta HMOVE

	clc

	;Y = 7 - abs(scorey)
	lda scorey
	bmi ScoreYNeg
	beq ScoreYZero

	;scorey > 0, burn scorey lines
	tay
ScoreYPosBurn
	sta WSYNC
	dey
	bne ScoreYPosBurn

ScoreYZero
	lda #8
	sbc scorey			;carry is clear
	bcs ScoreYDone

ScoreYNeg
	;negate scorey
	eor #$FF
	adc #1
	sta scoreyb
	lda scorey
	adc #7

ScoreYDone
	tay
	lda frame
	and #1
	sta HMCLR
	beq Call_ScoreKernelEven
	jsr ScoreKernelOdd
	jmp ContinueKernel

Call_ScoreKernelEven
	jsr ScoreKernelEven

ContinueKernel
	;decimal point
	lda #0
	sta NUSIZ1
	sta GRP0
	sta GRP1

	;Only show decimal point if scorey <= 0
	lda scorey
	cmp #1
	bpl NoDecimalPoint

	lda cenam1
	sta ENAM1

NoDecimalPoint
	lda state
	sta REFP1
	lda rabbref
	sta REFP0

	;position player
	lda xpos+1
	ldx #0
	jsr PositionObject

	;lots of time left here
	lda #0
	sta ENAM1
	sta NUSIZ0

	;burn lines for scorey < 0
	lda scoreyb
	beq PostScoreBurnDone

	tax
PostScoreBurn
	sta WSYNC
	dex
	bne PostScoreBurn

PostScoreBurnDone
	;position first bell
	lda bells
	ldx #1
	jsr PositionObject

	;also lots of time here

	sta WSYNC
	sta HMOVE

	;double width of GRP1 to match BELLW
	lda #5
	sta NUSIZ1

	;colors
	lda #PLAYCLR
	sta COLUP0
	lda #BELLCLR
	sta COLUP1

	lda #KLINES
	sta temp
	ldx #0
	lda #>Bird0
	sta temp+2
	sta HMCLR		;note: enough time should have passed, making it safe to HMCLR

	jmp MainKernel
MainKernel_Ret

	sta WSYNC

	;black line
	lda #0
	sta COLUBK
	sta COLUPF
	sta GRP0
	sta GRP1

	;white "mouse pointer"
	ldx #BLINES-1
Marker
	sta WSYNC
	lda #2
	sta ENAM0
	dex
	bne Marker

	;VBLANK
	lda #$82
	sta VBLANK
#if PAL
	ldx #OVERSCN
OverScanWait
	sta WSYNC
	lda #0
	sta ENAM0
	dex
	bne OverScanWait
#else
OverscanLogic
    lda #OVERS64
    sta TIM64T
    sta WSYNC
    lda #0
    sta ENAM0

    OVERSCAN_LOGIC

OverScanWait
    sta WSYNC
    lda INTIM
    bne OverScanWait
    sta WSYNC
#endif
	jmp MainLoop

;***********************
;** utility functions **
;***********************

;fibonacchi PRNG, with some upper bits being mixed it
;the adc #1 should guarantee that it never reaches a fixed point
prng
	adc rand
	adc rand+1
	sta rand
	adc rand+1
	adc #1
	adc rand
	sta rand+1
	rts

;shift bells according to belshft
ShiftBells
	lda belshft
	clc
	adc belstrt
	sta belstrt
	
	cmp #1
	bpl ShiftBells_Pos

	clc
	adc #BELLH+GAP
	sta belstrt
	
	;memmove(bells, bells+2, 2nbells-2)
	ldx #0
ShiftBells_Neg_Loop
	lda bells+2,X
	sta bells,X
	inx
	lda bells+2,X
	sta bells,X
	inx
	cpx #2*NBELLS-2
	bne ShiftBells_Neg_Loop

	;set last bell to invisible
	stx bells+NBELLS*2-1

	;bird moved if it's visible - update brdcntr in that case
	lda brdcntr
	bmi ShiftBells_Done
	sec
	sbc #2
	sta brdcntr
	rts

ShiftBells_Pos
	cmp #BELLH+GAP+1
	bmi ShiftBells_Done
	
	sec
	sbc #BELLH+GAP
	sta belstrt
	
	;memmove(bells+2, bells, 2*nbells-2);
	ldx #2*NBELLS-2
ShiftBells_Pos_Loop
	lda bells-1,X
	sta bells+1,X
	dex
	lda bells-1,X
	sta bells+1,X
	dex
	bne ShiftBells_Pos_Loop
	
	;bells[0..1] = InitBell() or InitBird() once every BRDFREQ bell if the game has started
	ldx #2
	lda state
	bpl InitBell
	lda brdcntr
	clc
	adc #2
	sta brdcntr
	beq InitBird
	cmp #NBELLS*2
	bmi InitBell
	lda #<(-2*BRDFREQ)
	sta brdcntr
	jmp InitBell

ShiftBells_Done
	rts

;init bell pointed to by X
;more specifically, init bell at bells[X-2..X-1]
InitBell
	;start on random frame
	jsr prng
	and #$3F
	ora #$80
	sta bells-1,X

	;xpos
	jsr prng

	;constrain to [0,BXMAX-BXMIN) via repeated subtraction
	;we always subtract at least once, so if prng returned zero we'd
	;STA #0 - (BXMAX-BXMIN) + BXMAX = BXMIN
	sec
InitBell_Constrain
	sbc #BXMAX-BXMIN
	bcs InitBell_Constrain

InitBell_Done
	adc #BXMAX
	sta bells-2,X
	rts

InitBird
	jsr InitBell
	lda bells-1,X
	ora #$C0
	sta bells-1,X
	rts

;Add score for touching a bell.
;The first bell is worth 10, the next 20, then 30 and so on.
ScoreBell
	sed

	lda bellscr+1
	cmp #$99
	beq ScoreBell_DoAdd
	lda score+SCORESZ-1
	cmp #SCORED+1
	bcs ScoreBell_DoAdd
	cmp #SCORED
	beq ScoreBell_IncOne

	lda #$10
	.byte $2C
ScoreBell_IncOne
	lda #$01
	clc
	adc bellscr
	sta bellscr
	lda #0
	adc bellscr+1
	sta bellscr+1

ScoreBell_DoAdd
	lda score
	clc
	adc bellscr
	sta score
	lda score+1
	adc bellscr+1
	sta score+1
	bcc CLD_RTS

	ldx #2
	clc
ScoreBell_DoCarry
	lda #1
	adc score,X			;we used to inc here, but that doesn't work
	sta score,X			;as expected in BCD mode
	bcc CLD_RTS			;Note: this used to be BNE, but Z isn't set sensefully in BCD mode
	inx
	cpx #SCORESZ-1
	bne ScoreBell_DoCarry
	beq ShiftScore

;Double the score
ScoreBird
	sed

	;The following is due to us not being able to just LDX #0
	;then loop until X == SCORESZ-1 since that could involve
	;a CPX #SCORESZ-1 which would clear our precious carry flag.
	;Hence we index using a "negative" X counting up to zero
	;by wrapping from the last page to the zero page.
	clc
	ldx #256-(SCORESZ-1)
ScoreBird_Loop
	lda $FF00+score+SCORESZ-1,X		;wastes three bytes of ROM and three cycles
	adc $FF00+score+SCORESZ-1,X		;but the net saving in ROM is worth it
	sta $FF00+score+SCORESZ-1,X
	inx
	bne ScoreBird_Loop
	bcc CLD_RTS

	;fall through to ShiftScore

;Shift mantissa one digit to the right and increase the exponent
;If the exponent is zero, then were going from integer to scientific
;mode which means we need to set the exponent to 8.
;Also shift bellscr one digit to the right.
ShiftScore
	ldy #4
ShiftScore_Loop
	lsr bellscr+1
	ror bellscr
	clc
	ldx #SCORESZ-1
ShiftScore_Loop2
	ror score-1,X
	dex
	bne ShiftScore_Loop2
	dey
	bne ShiftScore_Loop

	lda score+SCORESZ-2
	ora #$10
	sta score+SCORESZ-2

	lda score+SCORESZ-1
	bne ShiftScore_IncExp

	lda #SCORED-1

ShiftScore_IncExp
	clc
	adc #1
	sta score+SCORESZ-1

CLD_RTS
	cld
	rts

;Music code
HandleMusic
	ldx #2

HandleMusic_Loop
	;X is 2 the first time in here, 0 the second time
	;it gets halved for accessing inscnt0,X and AUD*0,X
	lda insptr0,X
	sta temp
	lda insptr0+1,X
	sta temp+1

	;halve X
	txa
	lsr
	tax

	ldy inscnt0,X
	beq HandleMusic_Silence

	lda (temp),Y
	sta AUDF0,X
	dey
	lda (temp),Y
	sta AUDC0,X
	lsr
	lsr
	lsr
	lsr
	dey
	.byte $2c
HandleMusic_Silence
	lda #0
	sta AUDV0,X
	sty inscnt0,X

	dex						;0 or -1
	bpl HandleMusic_Loop
	rts

;Setup the free:est channel
LoadFreeChannel
	lda inscnt0
	cmp inscnt1
	bpl LoadInstrument1		;channel 0 has more samples left than channel 1
							;meaning the second is more free to play

;Setup channel 0 to use instrument Y
;Affects: A, Y
LoadInstrument0
	lda Instruments,Y
	sta insptr0
	lda Instruments+1,Y
	sta insptr0+1
	ldy #0
	lda (insptr0),Y
	sta inscnt0
	rts

;Setup channel 1 to use instrument Y
;Affects: A, Y
LoadInstrument1
	lda Instruments,Y
	sta insptr1
	lda Instruments+1,Y
	sta insptr1+1
	ldy #0
	lda (insptr1),Y
	sta inscnt1
	rts

;Called when the game starts
;In other words, when state transitions from #$40 to #$C0
StartGame
	;init scores
	lda #0
	ldx #SCORESZ
StartGame_ClearScore
	sta score-1,X
	dex
	bne StartGame_ClearScore

	sta bellscr
	sta bellscr+1

	lda #<(-2*BRDFREQ)
	sta brdcntr

	;set b7 of state
	lda state
	ora #$80
	sta state
	rts

;Called when the game ends
;In other words, when state == #$80 (game started but on ground)
EndGame
	;scroll to reveal hiscore
	lda #0
	sta scorey
	sta scorcnt
	lda #HOLD
	sta holdcnt
	lda #NOHISCRLINE
	sta curline

	;update hiscore
	;simply scan both score and hiscore from MSB to LSB to figure out whether score is larger
	ldx #SCORESZ+1
EndGame_CheckHiscore
	dex
	beq EndGame_NoHiscore
	lda score-1,X
	cmp hiscore-1,X
	bcc EndGame_NoHiscore		;score < hiscore
	beq EndGame_CheckHiscore	;score+X == hiscore+X, check next byte
	bcs EndGame_HiScore			;score > hiscore

EndGame_HiScore
	lda #HISCORELINE
	sta curline

	ldx #SCORESZ
EndGame_HiScore_Loop
	lda score-1,X
	sta hiscore-1,X
	dex
	bne	EndGame_HiScore_Loop

	;play hiscore sound
	ldy #4
	.byte $2C
EndGame_NoHiscore
	;play no-hiscore sound
	ldy #6
	lda #0
	sta state

	;make hiscore display immediately
	lda #$80
	sta frame

	jmp LoadFreeChannel		;jump, the function will then rts
							;this save one byte of ROM

	;Positions object X to horizontal position in accumulator
	; 0 = P0
	; 1 = P1
	; 2 = M0
	; 3 = M1
	; 4 = BL
	;
	;We purposefully burn enough cycles when we're done so that
	;this function consumes at least one full scanline
PositionObject
	clc
	sta temp
	sta WSYNC
DivideLoop
	sbc #15
	bcs DivideLoop

	;The might become unaligned without us noticing, causing odd
	;mistimings in the display code.
	;Print the page difference so this can be more easily spotted.
	echo "DivideLoop page difference:", (>DivideLoop - >*)

	eor #7
	asl
	asl
	asl
	asl
	sta HMP0,X
	sta RESP0,X

	lda #175		;large enough that it should guarantee we'll waste enough time
	sbc temp
DivideLoop2
	sbc #15
	bcs DivideLoop2

	;Same as for DivideLoop
	echo "DivideLoop2 page difference:", (>DivideLoop2 - >*)
	rts

	;the only reason I've put this here is that putting these graphics in the header
	;shifts the alignment enough to cause the score kernel to be misaligned
	include starfieldgfx.h

	;ditto for this table
	;the entries below delimit which frames are used below which velocities
RabbitFrames
							;frame 0 (implied)
	.word	200				;frame 1
	.word	150				;frame 2
	.word	100				;frame 3
	.word	50				;frame 4
	.word	-50				;frame 5
	.word	-150			;frame 6
	.word	-200			;frame 7
	.word	-250			;frame 8
							;frames 9 and 10 are used when yvel+1 == FALLMIN  (ears flutter)

PickJumpFrame
	;pick sprite based on vertical speed
	lda yvel+1
	cmp #<FALLMIN
	bne NegativeYvel_DoLoop

	;flutter
	lda frame
	asl
	and #8
	clc
	adc #<Rabbit9
	sta p0ptr
	rts

NegativeYvel_DoLoop
	;four entries in table
	ldx #13
NegativeYvel_Loop
	lda yvel+1
	cmp RabbitFrames,X
	beq NegativeYvel_TestLow
	bmi NegativeYvel_Done
	bpl NegativeYvel_Next

NegativeYvel_TestLow
	lda yvel
	cmp RabbitFrames-1,X
	bcc NegativeYvel_Done

NegativeYvel_Next
	dex
	dex
	bpl NegativeYvel_Loop

NegativeYvel_Done
	inx
	txa
	asl
	asl
	clc
	adc #<Rabbit0
	sta p0ptr
	rts

PickHopFrame
	lda frame
	and #$03
	bne PickHopFrame_Done

	;increase p0ptr, keep it between Hop0 and Hop5
	clc
	lda p0ptr
	adc #8
	cmp #<Hop0+8*HOPFRMS
	bcs PickHopFrame_Reset
	cmp #<Hop0
	bcc PickHopFrame_Reset
	.byte $2C
PickHopFrame_Reset
	lda #<Hop0
	sta p0ptr

PickHopFrame_Done
	rts

PickSitFrame
	lda frame
	and #$07
	bne PickSitFrame_Done

	lda p0ptr
	cmp #<Sit0
	bcc PickSitFrame_Reset
	cmp #<Sit0+8*(SITFRMS-1)
	beq PickSitFrame_Done
	clc
	adc #8
	.byte $2C
PickSitFrame_Reset
	lda #<Sit0
	sta p0ptr

PickSitFrame_Done
	rts

	echo "ROM:", ($FFFC - *), "bytes left"

	org $FFFC
	.word Start
	.word Start
